"""
Wrapper for subscript type annotations, like `List[str]`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Generator, Iterable, Iterator
from typing import Self

from mypy_boto3_builder.exceptions import TypeAnnotationError
from mypy_boto3_builder.import_helpers.import_record import ImportRecord
from mypy_boto3_builder.type_annotations.fake_annotation import FakeAnnotation
from mypy_boto3_builder.type_annotations.type_parent import TypeParent


class TypeSubscript(TypeParent):
    """
    Wrapper for subscript type annotations, like `List[str]`.

    Arguments:
        parent -- Parent type annotation.
        children -- Children type annotations.
    """

    def __init__(
        self,
        parent: FakeAnnotation,
        children: Iterable[FakeAnnotation] = (),
    ) -> None:
        self.parent: FakeAnnotation = parent
        self.children: list[FakeAnnotation] = list(children)

    def render(self) -> str:
        """
        Render type annotation to a valid Python code for local usage.

        Returns:
            A string with a valid type annotation.
        """
        result = self.parent.render()
        if self.children:
            children = ", ".join([i.render() for i in self.children])
            result = f"{result}[{children}]"

        return result

    def get_import_records(self) -> set[ImportRecord]:
        """
        Get all import records required for using type annotation.
        """
        result: set[ImportRecord] = set()
        result.update(self.parent.get_import_records())
        for child in self.children:
            result.update(child.get_import_records())
        return result

    def iterate_types(self) -> Generator[FakeAnnotation]:
        """
        Extract type annotations from children.
        """
        yield from self.parent.iterate_types()
        for child in self.children:
            yield from child.iterate_types()

    def add_child(self, child: FakeAnnotation) -> None:
        """
        Add new child to Substcript.
        """
        self.children.append(child)

    def is_dict(self) -> bool:
        """
        Whether subscript parent is Dict.
        """
        return self.parent.is_dict()

    def is_list(self) -> bool:
        """
        Whether subscript parent is List.
        """
        return self.parent.is_list()

    def __copy__(self) -> Self:
        """
        Create a copy of type annotation wrapper.
        """
        return self.__class__(
            parent=self.parent,
            children=list(self.children),
        )

    def get_local_types(self) -> list[FakeAnnotation]:
        """
        Get internal types generated by builder.
        """
        result: list[FakeAnnotation] = []
        for child in self.children:
            result.extend(child.get_local_types())
        return result

    def iterate_direct_type_annotations(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over a parent and then children type annotations.
        """
        yield self.parent
        yield from self.children

    def get_children_types(self) -> set[FakeAnnotation]:
        """
        Extract required type annotations from attributes.
        """
        result: set[FakeAnnotation] = set()
        for child in self.children:
            result.update(child.iterate_types())
        return result

    def replace_child(self, child: FakeAnnotation, new_child: FakeAnnotation) -> Self:
        """
        Replace child type annotation with a new one.
        """
        if self.parent is child:
            self.parent = new_child
            return self

        if child not in self.children:
            raise TypeAnnotationError(f"Child not found: {child}")

        index = self.children.index(child)
        self.children[index] = new_child
        return self

    def iterate_children(self) -> Iterator[FakeAnnotation]:
        """
        Iterate over children.
        """
        yield from self.children



"""
Ruff formatter.

Copyright 2024 Vlad Emelianov
"""

import json
import subprocess
import sys
import tempfile
from collections.abc import Iterable, Sequence
from pathlib import Path

from mypy_boto3_builder.constants import LINE_LENGTH, SUPPORTED_PY_VERSIONS
from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.utils.path import print_path


class RuffError(RuntimeError):
    """
    Ruff error.
    """


class RuffFormatter:
    """
    Ruff formatter.
    """

    def __init__(
        self,
        known_first_party: Sequence[str] = (),
        known_third_party: Sequence[str] = (),
    ) -> None:
        self.logger = get_logger()
        self._target_version = self._get_target_version()
        self._known_first_party = list(known_first_party)
        self._known_third_party = [i for i in known_third_party if i not in self._known_first_party]

    @staticmethod
    def _get_target_version() -> str:
        min_version = min(v for v in SUPPORTED_PY_VERSIONS if len(v) > 1)
        return f"py{min_version[0]}{min_version[1]}"

    def format_python(self, paths: Sequence[Path]) -> None:
        """
        Format python files with `ruff`.

        Arguments:
            path -- Target path.
        """
        self._sort_imports(paths)
        self._run_format(paths)

    def _get_config_cli(self) -> list[str]:
        overrides = [
            f'target-version = "{self._target_version}"',
            f"line-length = {LINE_LENGTH}",
            'format.quote-style = "double"',
            'format.line-ending = "lf"',
            f"lint.isort.known-first-party = {json.dumps(self._known_first_party)}",
            f"lint.isort.known-third-party = {json.dumps(self._known_third_party)}",
        ]
        result: list[str] = []
        for option in overrides:
            result.extend(["--config", option])

        return result

    def _run_check(self, paths: Sequence[Path], rules: Sequence[str]) -> None:
        cmd = (
            sys.executable,
            "-m",
            "ruff",
            "check",
            "--target-version",
            self._target_version,
            *self._get_config_cli(),
            "--select",
            ",".join(rules),
            "--fix",
            "--isolated",
            *(path.as_posix() for path in paths),
        )
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.warning(
                f"Ruff check failed for paths {[print_path(path) for path in paths]}",
            )
            self.logger.warning(" ".join(cmd))
            self.logger.warning(e.output.decode())
            raise RuffError(f"Ruff check failed with status {e.returncode}") from None

    def _sort_imports(self, paths: Sequence[Path]) -> None:
        """
        Sort imports.
        """
        self._run_check(paths, ("I",))

    def _run_format(self, paths: Sequence[Path]) -> None:
        """
        Format python code with `ruff`.
        """
        cmd = (
            sys.executable,
            "-m",
            "ruff",
            "format",
            "--target-version",
            self._target_version,
            *self._get_config_cli(),
            "--isolated",
            *(path.as_posix() for path in paths),
        )
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            self.logger.warning(
                f"Ruff format failed for paths {[print_path(path) for path in paths]}",
            )
            self.logger.warning(" ".join(cmd))
            self.logger.warning(e.output.decode())
            raise RuffError(f"Ruff format failed: {e.output}, paths: {paths}") from None

    def format_strings(self, codes: Iterable[str]) -> list[str]:
        """
        Format python code as strings.
        """
        with tempfile.TemporaryDirectory() as dir_name:
            paths: list[Path] = []
            for index, code in enumerate(codes):
                file_path = Path(dir_name) / f"temp_{index}.py"
                file_path.write_text(code)
                paths.append(file_path)

            self._run_format(paths)
            return [path.read_text().rstrip("\n") for path in paths]

    def format_markdown(self, text: str) -> str:
        """
        Format python codeblocks in markdown.
        """
        blocks = text.split("\n```")
        format_blocks: list[str] = []
        format_block_indices: list[int] = []
        for index, block in enumerate(blocks):
            if block.startswith("python"):
                format_blocks.append(block)
                format_block_indices.append(index)

        if format_blocks:
            for index, formatted_block in enumerate(self.format_strings(format_blocks)):
                block_index = format_block_indices[index]
                blocks[block_index] = formatted_block
        return "\n```".join(blocks)



"""
Collection of TypeUnions added by boto3.

Copyright 2024 Vlad Emelianov
"""

from botocore.response import StreamingBody

from mypy_boto3_builder.import_helpers.import_helper import Import
from mypy_boto3_builder.type_annotations.external_import import ExternalImport
from mypy_boto3_builder.type_annotations.type import Type
from mypy_boto3_builder.type_annotations.type_subscript import TypeSubscript
from mypy_boto3_builder.type_annotations.type_union import TypeUnion
from mypy_boto3_builder.type_maps.typed_dicts import (
    AttributeValueTypeDef,
    CopySourceTypeDef,
    PolicyDocumentDictTypeDef,
)

StreamingBodyType = ExternalImport.from_class(StreamingBody)

ConditionBaseImportTypeDef = TypeUnion(
    name="ConditionBaseImportTypeDef",
    children=(
        Type.str,
        ExternalImport(Import.boto3 + "dynamodb" + "conditions", "ConditionBase"),
    ),
)

CopySourceOrStrTypeDef = TypeUnion(
    name="CopySourceOrStrTypeDef",
    children=(Type.str, CopySourceTypeDef),
)

BlobTypeDef = TypeUnion(
    name="BlobTypeDef",
    children=(Type.str, Type.bytes, Type.IOAny, StreamingBodyType),
)

TimestampTypeDef = TypeUnion(name="TimestampTypeDef", children=(Type.datetime, Type.str))

FileobjTypeDef = TypeUnion(
    name="FileobjTypeDef",
    children=[Type.IOAny, StreamingBodyType],
)

VerifyTypeDef = TypeUnion((Type.bool, Type.str, Type.none))
PresignedPostConditionsTypeDef = TypeUnion((Type.ListAny, Type.DictStrAny))

# FIXME: a hack to avoid cicular TypedDict in dynamodb package
TableAttributeValueTypeDef = TypeUnion(
    name="TableAttributeValueTypeDef",
    children=(
        Type.bytes,
        Type.bytearray,
        Type.str,
        Type.int,
        Type.Decimal,
        Type.bool,
        TypeSubscript(Type.Set, [Type.int]),
        TypeSubscript(Type.Set, [Type.Decimal]),
        TypeSubscript(Type.Set, [Type.str]),
        TypeSubscript(Type.Set, [Type.bytes]),
        TypeSubscript(Type.Set, [Type.bytearray]),
        Type.SequenceAny,
        Type.MappingStrAny,
        Type.none,
    ),
)

UniversalAttributeValueTypeDef = TypeUnion(
    name="UniversalAttributeValueTypeDef",
    children=(
        AttributeValueTypeDef,
        *TableAttributeValueTypeDef.children,
    ),
)

PolicyDocumentTypeDef = TypeUnion(
    name="PolicyDocumentTypeDef",
    children=(Type.str, PolicyDocumentDictTypeDef),
)



"""
Parser that produces `structures.ServicePackage`.

Copyright 2024 Vlad Emelianov
"""

from collections.abc import Iterable

from mypy_boto3_builder.logger import get_logger
from mypy_boto3_builder.package_data import BasePackageData
from mypy_boto3_builder.parsers.client_parser import ClientParser
from mypy_boto3_builder.parsers.service_resource_parser import ServiceResourceParser
from mypy_boto3_builder.parsers.shape_parser import ShapeParser
from mypy_boto3_builder.service_name import ServiceName, ServiceNameCatalog
from mypy_boto3_builder.structures.method import Method
from mypy_boto3_builder.structures.packages.service_package import ServicePackage
from mypy_boto3_builder.structures.paginator import Paginator
from mypy_boto3_builder.structures.waiter import Waiter
from mypy_boto3_builder.type_annotations.type_def_sortable import TypeDefSortable
from mypy_boto3_builder.type_maps.typed_dicts import CloudwatchEventTypeDef
from mypy_boto3_builder.utils.strings import xform_name
from mypy_boto3_builder.utils.type_def_sorter import TypeDefSorter


class ServicePackageParser:
    """
    Parser that produces `structures.ServicePackage`.

    Arguments:
        service_name -- Target service name.
        package_data -- Package data.
        version -- Package version.

    Returns:
        ServiceModule structure.
    """

    def __init__(
        self,
        service_name: ServiceName,
        package_data: BasePackageData,
        version: str,
    ) -> None:
        self.service_name = service_name
        self.package_data = package_data
        self.version = version
        self.shape_parser = ShapeParser(self.service_name)
        self._logger = get_logger()

    def parse(self) -> ServicePackage:
        """
        Extract all data from boto3 service package.
        """
        result = self._parse_service_package()
        result.waiters.extend(self._parse_waiters())
        result.waiters.sort()

        result.paginators.extend(self._parse_paginators())
        result.paginators.sort()

        result.client.methods.extend(
            self._get_extra_client_methods(result.paginators, result.waiters),
        )

        self.shape_parser.fix_typed_dict_names()
        self.shape_parser.convert_input_arguments_to_unions(result.iterate_methods())

        type_defs = result.extract_type_defs()

        # Add hardcoded CloudwatchEvent type definitions to cloudwatch service
        if self.service_name == ServiceNameCatalog.cloudwatch:
            type_defs.add(CloudwatchEventTypeDef)

        result.type_defs = self._get_sorted_type_defs(type_defs)
        result.literals = result.extract_literals()
        result.validate()
        result.calculate_install_requires()

        return result

    def _parse_service_package(self) -> ServicePackage:
        service_resource_parser = ServiceResourceParser(
            self.service_name,
            self.shape_parser,
        )
        client_parser = ClientParser(
            self.service_name,
            self.shape_parser,
        )
        client = client_parser.parse()
        service_resource = service_resource_parser.parse_service_resource()

        return ServicePackage(
            data=self.package_data,
            service_name=self.service_name,
            client=client,
            service_resource=service_resource,
            version=self.version,
        )

    def _parse_waiters(self) -> list[Waiter]:
        waiters: list[Waiter] = []
        for waiter_name in self.shape_parser.get_waiter_names():
            self._logger.debug(f"Parsing Waiter {waiter_name}", tags=waiter_name)
            waiter = Waiter(
                name=f"{waiter_name}Waiter",
                waiter_name=waiter_name,
                attribute_name=xform_name(waiter_name),
                service_name=self.service_name,
            )
            wait_method = self.shape_parser.get_wait_method(waiter_name)
            waiter.methods.append(wait_method)
            waiters.append(waiter)

        return waiters

    def _parse_paginators(self) -> list[Paginator]:
        result: list[Paginator] = []
        for paginator_name in self.shape_parser.get_paginator_names():
            self._logger.debug(f"Parsing Paginator {paginator_name}", tags=paginator_name)
            operation_name = xform_name(paginator_name)
            paginator_record = Paginator(
                name=f"{paginator_name}Paginator",
                paginator_name=paginator_name,
                operation_name=operation_name,
                service_name=self.service_name,
                return_type=self.shape_parser.get_paginator_subscript(paginator_name),
            )

            paginate_method = self.shape_parser.get_paginate_method(paginator_name)
            paginator_record.methods.append(paginate_method)
            result.append(paginator_record)

        return result

    def _get_extra_client_methods(
        self,
        paginators: list[Paginator],
        waiters: list[Waiter],
    ) -> list[Method]:
        result: list[Method] = []
        for paginator in paginators:
            method = paginator.get_client_method()
            if len(paginators) == 1:
                method.decorators.clear()
            result.append(method)

        for waiter in waiters:
            method = waiter.get_client_method()
            if len(waiters) == 1:
                method.decorators.clear()
            result.append(method)

        return result

    def _get_sorted_type_defs(self, type_defs: Iterable[TypeDefSortable]) -> list[TypeDefSortable]:
        type_def_sorter = TypeDefSorter(type_defs)
        return type_def_sorter.sort()



"""
Output type choice for CLI.

Copyright 2024 Vlad Emelianov
"""

from enum import Enum


class OutputType(Enum):
    """
    Output type choice for CLI.
    """

    package = "package"
    wheel = "wheel"
    sdist = "sdist"
    installed = "installed"

    def is_installed(self) -> bool:
        """
        Whether output type is a ready-to-use directory.
        """
        return self is OutputType.installed

    def is_packaged(self) -> bool:
        """
        Whether output type is packaged for installation.
        """
        return self is OutputType.wheel or self is OutputType.sdist

    def is_preserved(self) -> bool:
        """
        Check if output is preserved.
        """
        return not self.is_packaged()



"""
Generator for mypy-boto3 package.

Copyright 2025 Vlad Emelianov
"""

from typing import NoReturn

from mypy_boto3_builder.constants import TemplatePath
from mypy_boto3_builder.exceptions import AlreadyPublishedError
from mypy_boto3_builder.generators.base_generator import BaseGenerator
from mypy_boto3_builder.package_data import Boto3StubsPackageData, MypyBoto3PackageData
from mypy_boto3_builder.parsers.mypy_boto3_package import parse_mypy_boto3_package
from mypy_boto3_builder.structures.packages.mypy_boto3_package import MypyBoto3Package
from mypy_boto3_builder.structures.packages.service_package import ServicePackage


class MypyBoto3Generator(BaseGenerator):
    """
    Generator for mypy-boto3 package.
    """

    service_package_data = Boto3StubsPackageData()

    def generate_stubs(self) -> MypyBoto3Package | None:
        """
        Generate main stubs.
        """
        package_data = MypyBoto3PackageData()
        try:
            version = self._get_package_build_version(package_data.pypi_name)
        except AlreadyPublishedError:
            return None

        self._log_generate(package_data.pypi_name, version)
        package = parse_mypy_boto3_package(service_names=self.main_service_names, version=version)
        self.package_writer.write_package(
            package,
            template_path=TemplatePath.mypy_boto3,
        )
        return package

    def generate_stubs_lite(self) -> NoReturn:
        """
        Not implemented.
        """
        raise NotImplementedError("Should never be called")

    def generate_full_stubs(self) -> NoReturn:
        """
        Not implemented.
        """
        raise NotImplementedError("Should never be called")

    def generate_custom_stubs(self) -> NoReturn:
        """
        Not implemented.
        """
        raise NotImplementedError("Should never be called")

    def generate_docs(self) -> None:
        """
        Not implemented.
        """
        raise NotImplementedError("Should never be called")

    def _get_postprocessor(self, service_package: ServicePackage) -> NoReturn:
        """
        Not implemented.
        """
        raise NotImplementedError("Should never be called")



"""
Jinja2 `Environment` manager.

Copyright 2024 Vlad Emelianov
"""

from pathlib import Path
from typing import Any, ClassVar

from jinja2.environment import Environment, Template
from jinja2.loaders import FileSystemLoader
from jinja2.runtime import StrictUndefined

from mypy_boto3_builder.constants import TEMPLATES_PATH
from mypy_boto3_builder.exceptions import JinjaManagerError
from mypy_boto3_builder.utils.strings import escape_md

__all__ = ["JinjaManager"]


class JinjaManager:
    """
    Jinja2 `Environment` manager.
    """

    _environment: ClassVar = Environment(
        loader=FileSystemLoader(TEMPLATES_PATH.as_posix()),
        undefined=StrictUndefined,
    )
    _template_cache: ClassVar[dict[Path, Template]] = {}

    def __init__(self) -> None:
        self._environment.filters["escape_md"] = escape_md

    @classmethod
    def update_globals(cls, **kwargs: Any) -> None:  # noqa: ANN401
        """
        Update global variables in `jinja2.Environment`.

        Arguments:
            kwargs -- Globals to set.
        """
        cls._environment.globals.update(kwargs)

    def get_template(self, template_path: Path) -> Template:
        """
        Get `jinja2.Template`.
        """
        if template_path.is_absolute():
            template_path = template_path.relative_to(TEMPLATES_PATH)

        if template_path in self._template_cache:
            return self._template_cache[template_path]

        template_full_path = TEMPLATES_PATH / template_path

        if not template_full_path.exists():
            raise JinjaManagerError(f"Template {template_full_path} not found")

        template: Template = self._environment.get_template(template_path.as_posix())
        self._template_cache[template_path] = template

        return template



"""
Type annotations for aioboto3.resources.response module.

Copyright 2025 Vlad Emelianov
"""

from typing import Any

from boto3.resources.base import ServiceResource
from boto3.resources.response import RawHandler, ResourceHandler

class AIOResourceHandler(ResourceHandler):
    async def __call__(  # type: ignore[override]
        self, parent: ServiceResource, params: dict[str, Any], response: dict[str, Any]
    ) -> ServiceResource | list[ServiceResource]: ...

class AIORawHandler(RawHandler):
    async def __call__(  # type: ignore[override]
        self, parent: ServiceResource, params: dict[str, Any], response: dict[str, Any]
    ) -> dict[str, Any]: ...



"""
Type annotations for aioboto3.resources.factory module.

Copyright 2025 Vlad Emelianov
"""

import logging
from typing import Any

from aioboto3.resources.base import AIOBoto3ServiceResource
from boto3.resources.factory import ResourceFactory
from boto3.utils import ServiceContext
from botocore.hooks import BaseEventHooks

logger: logging.Logger

class AIOBoto3ResourceFactory(ResourceFactory):
    def __init__(self, emitter: BaseEventHooks) -> None: ...
    async def load_from_definition(  # type: ignore[override]
        self,
        resource_name: str,
        single_resource_json_definition: Any,
        service_context: ServiceContext,
    ) -> AIOBoto3ServiceResource: ...
